#pragma once

#include <cstddef>

// these are generated by ../../misc/xxd.py

#ifdef __cplusplus
extern "C" {
#endif

// ../doc/murphi2c.1
extern const unsigned char doc_murphi2c_1[];
extern const size_t doc_murphi2c_1_len;

// ../resources/c_prefix.c
extern const unsigned char resources_c_prefix_c[];
extern const size_t resources_c_prefix_c_len;

// ../resources/h_prefix.h
extern const unsigned char resources_h_prefix_h[];
extern const size_t resources_h_prefix_h_len;

#ifdef __cplusplus
}
#endif
