#pragma once

#include <cstddef>

// these are generated by ../../misc/xxd.py

#ifdef __cplusplus
extern "C" {
#endif

// ../doc/murphi2murphi.1
extern const unsigned char doc_murphi2murphi_1[];
extern const size_t doc_murphi2murphi_1_len;

#ifdef __cplusplus
}
#endif
